"""Authentication helper functions
"""
import os
from flask import request
from .jwt import jwt_manager
from flask_jwt_extended import jwt_required, get_jwt_identity


# Accept user name passed in Basic Auth header
# (password has to checkded before!)
ALLOW_BASIC_AUTH_USER = os.environ.get('ALLOW_BASIC_AUTH_USER', 'False') \
    .lower() in ('t', 'true')


def auth_manager(app, api=None):
    """Authentication setup for Flask app"""
    # Setup the Flask-JWT-Extended extension
    return jwt_manager(app, api)


def optional_auth(fn):
    """Authentication view decorator"""
    return jwt_required(optional=True)(fn)


def get_identity():
    """Get identity (username oder dict with username and groups)"""
    return get_jwt_identity()


def get_username(identity):
    """Get username"""
    if identity:
        if isinstance(identity, dict):
            username = identity.get('username')
        else:
            # identity is username
            username = identity
    else:
        username = None
    return username


def get_groups(identity):
    """Get user groups"""
    groups = []
    if identity:
        if isinstance(identity, dict):
            groups = identity.get('groups', [])
            group = identity.get('group')
            if group:
                groups.append(group)
    return groups


def get_auth_user():
    """Get identity or optional pre-authenticated basic auth user"""
    identity = get_identity()
    if not identity and ALLOW_BASIC_AUTH_USER:
        auth = request.authorization
        if auth:
            # We don't check password, already authenticated!
            identity = auth.username
    return identity



import os
import re
from flask import json, safe_join


class RuntimeConfig:
    '''Runtime configuration helper class
    '''

    @staticmethod
    def config_file_path(service, tenant):
        """Return path to permissions JSON file for a tenant.

        :param str servcie: Service name
        :param str tenant: Tenant ID
        """
        config_path = os.environ.get('CONFIG_PATH', 'config')
        filename = '%sConfig.json' % service
        return safe_join(config_path, tenant, filename)

    def __init__(self, service, logger):
        self.service = service
        self.logger = logger
        self.config = None

    def read_config(self, tenant):
        """Read service config for a tenant from a JSON file.

        :param str tenant: Tenant ID
        """
        runtime_config_path = RuntimeConfig.config_file_path(
            self.service, tenant
        )
        self.logger.info(
            "Reading runtime config '%s'" % runtime_config_path
        )
        try:
            with open(runtime_config_path, encoding='utf-8') as fh:
                data = fh.read()
                # Replace env variables
                dataout = ENVVAR_PATTERN.sub(envrepl, data)
                self.config = json.loads(dataout)
        except Exception as e:
            self.logger.error(
                "Could not load runtime config '%s':\n%s" %
                (runtime_config_path, e)
            )
            self.config = {}
        return self

    def tenant_config(self, tenant):
        return self.read_config(tenant)

    def get(self, name, default=None):
        val = self.config.get('config', {}).get(name, default)
        # Optional override from env var
        envval = os.environ.get(name.upper())
        if envval is not None:
            # Convert from string
            if val is None:
                # unkown type --> no conversion
                val = envval
            elif type(val) in (list, dict):
                val = json.loads(envval)
            else:
                val = type(val)(envval)

        return val

    def resources(self):
        return self.config.get('resources')

    def resource(self, name):
        return self.config.get('resources', {}).get(name)


ENVVAR_PATTERN = re.compile(r'\$\$(\w+)\$\$')


def envrepl(match):
    name = match.group(1)
    val = os.environ.get(name, '')
    return val



import os
import datetime
from flask_jwt_extended import JWTManager, unset_jwt_cookies
from flask import make_response, Response


class JwtErrorHandlerProxy:
    def __init__(self, api):
        self.api = api

    def errorhandler(self, errortype):
        return lambda jwt_callback: self.api.errorhandler(errortype)(lambda error: self.jwt_callback_wrapper(jwt_callback, error))

    def jwt_callback_wrapper(self, jwt_callback, error):
        result = jwt_callback(error)
        # flask_restx error handler expects a plain object with an 'message' field
        # See flask_restx/api.py@handle_error
        if isinstance(result[0], Response):
            return {"message": result[0].json["msg"]}, result[1]
        return result


def jwt_manager(app, api=None):
    """Setup Flask-JWT-Extended extension for services
       with authenticated access"""
    # https://flask-jwt-extended.readthedocs.io/en/stable/options
    app.config['JWT_TOKEN_LOCATION'] = ['headers', 'cookies']
    app.config['JWT_ACCESS_COOKIE_NAME'] = os.environ.get(
        'JWT_ACCESS_COOKIE_NAME', 'access_token_cookie')
    app.config['JWT_COOKIE_CSRF_PROTECT'] = bool(os.environ.get(
        'JWT_COOKIE_CSRF_PROTECT', False))
    app.config['JWT_SECRET_KEY'] = os.environ.get(
        'JWT_SECRET_KEY', os.urandom(24))

    jwt = JWTManager(app)

    if api:
        api.__jwt_error_handler_proxy = JwtErrorHandlerProxy(api)

        # Delegate error handlers to flask_restx because of
        # https://github.com/vimalloc/flask-jwt-extended/issues/86
        jwt._set_error_handler_callbacks(api.__jwt_error_handler_proxy)

        @api.errorhandler
        def restplus_error_handler(error):
            # JWT error handler will be called afterwards
            return {}

    @jwt.expired_token_loader
    def handle_expired_token():
        # http://flask-jwt-extended.readthedocs.io/en/latest/changing_default_behavior.html
        # resp = redirect('/auth/login')
        # Automatic re-login does't work with SAML, so we prepare
        # for manual re-login
        resp = make_response()
        unset_jwt_cookies(resp)
        return resp

    return jwt



from flask_restx import Api as BaseApi
from collections import OrderedDict
from werkzeug.datastructures import MultiDict
from flask_restx.reqparse import Argument


class Api(BaseApi):
    """Custom Flask-RESTPlus Api subclass for overriding default root route

    NOTE: endpoint of route '/' must be named 'root'::

    >>> @api.route('/', endpoint='root')

    see also https://github.com/noirbizarre/flask-restplus/issues/247
    """
    def _register_doc(self, app_or_blueprint):
        if self._add_specs and self._doc:
            # Register documentation before root if enabled
            app_or_blueprint.add_url_rule(self._doc, 'doc', self.render_doc)
        # skip default root route
        # app_or_blueprint.add_url_rule(self.prefix or '/', 'root',
        #                               self.render_root)

    def create_model(self, name, fields):
        """Helper for creating api models with ordered fields

        :param str name: Model name
        :param list fields: List of tuples containing ['field name', <type>]
        """
        return create_model(self, name, fields)


def create_model(api, name, fields):
    """Helper for creating api models with ordered fields

    :param flask_restx.Api api: Flask-RESTX Api object
    :param str name: Model name
    :param list fields: List of tuples containing ['field name', <type>]
    """
    d = OrderedDict()
    for field in fields:
        d[field[0]] = field[1]
    return api.model(name, d)


class CaseInsensitiveMultiDict(MultiDict):
    """ A MultiDict subclass to use with RequestParser which is
        case-insensitives for query parameter key names
    """
    def __init__(self, mapping=None):
        super().__init__(mapping)
        # map lowercase keys to the real keys
        self.lower_key_map = {key.lower(): key for key in self}

    def __contains__(self, key):
        return key.lower() in self.lower_key_map

    def getlist(self, key):
        return super().getlist(self.lower_key_map.get(key.lower()))

    def pop(self, key):
        return super().pop(self.lower_key_map.get(key.lower()))


class CaseInsensitiveArgument(Argument):
    def source(self, request):
        return CaseInsensitiveMultiDict(super().source(request))



from copy import deepcopy
import os

from flask import json, safe_join
from .auth import get_username, get_groups


class PermissionsReader():
    """PermissionsReader helper class

    Read users, groups, roles and permissions for a tenant from a JSON file.
    Provides helper methods for collectiong permissions.
    """

    # name of public role
    PUBLIC_ROLE_NAME = 'public'

    @staticmethod
    def permissions_file_path(tenant):
        """Return path to permissions JSON file for a tenant.

        :param str tenant: Tenant ID
        """
        config_path = os.environ.get('CONFIG_PATH', 'config')
        return safe_join(config_path, tenant, 'permissions.json')

    def __init__(self, tenant, logger):
        """Constructor

        :param str tenant: Tenant ID
        :param Logger logger: Application logger
        """
        self.tenant = tenant
        self.logger = logger
        self.permissions = self.load_permissions()

    def read_permissions(self):
        """Read permissions for a tenant from a JSON file."""
        permissions = {}

        permissions_path = PermissionsReader.permissions_file_path(self.tenant)
        self.logger.info("Reading permissions '%s'" % permissions_path)
        try:
            with open(permissions_path, encoding='utf-8') as fh:
                permissions = json.load(fh)
        except Exception as e:
            self.logger.error(
                "Could not load permissions '%s':\n%s" %
                (permissions_path, e)
            )
            raise e
        # TODO: validate permissions schema

        return permissions

    def load_permissions(self):
        """Load users, groups, roles and permissions.

        Returns lookup dict as:
            {
                users: {
                    <user>: [<role>]
                }
                groups: {
                    <group>: [<role>]
                },
                roles: {
                    <role>: <permissions{}>
                }
            }
        """
        permissions = self.read_permissions()

        # transform raw permissions to lookup dict

        # detect permissions schema type
        is_unified = (
            'dataproducts' in permissions and 'common_resources' in permissions
        )

        resources_lookup = {}
        if is_unified:
            # create lookup for resources
            for resource in permissions.get('dataproducts', []):
                resources_lookup[resource.get('name')] = resource

        # collect group roles
        groups = {}
        for group in permissions.get('groups', []):
            groups[group['name']] = group['roles'] or []

        # collect user roles
        users = {}
        for user in permissions.get('users', []):
            user_roles = []
            # direct roles
            for role in user.get('roles', []) or []:
                user_roles.append(role)
            # roles from groups
            for group in user['groups'] or []:
                user_roles.extend(groups.get(group, []))
            # assign unique sorted roles
            users[user['name']] = sorted(list(set(user_roles)))

        # collect role permissions
        roles = {}
        for role in permissions.get('roles', []):
            if not is_unified:
                roles[role['role']] = role['permissions']
            else:
                roles[role['role']] = self.expand_unified_permissions(
                    role['permissions'], deepcopy(resources_lookup), permissions
                )

        return {
            'users': users,
            'groups': groups,
            'roles': roles
        }

    def expand_unified_permissions(self, role_permissions, resources_lookup,
                                   permissions):
        """Return full resource permissions expanded from unified permissions.

        :param obj role_permissions: Unified permissions for role
        :param obj resources_lookup: Lookup for resources with sublayers or
                                     attributes
        :param obj permissions: Permissions from JSON
        """
        full_permissions = {}

        wms_name = permissions.get('wms_name', '')
        wfs_name = permissions.get('wfs_name', '')
        # common resources for internal print layers, background layers,
        #   print templates and default solr facets
        common_resources = permissions.get('common_resources', [])

        wms_layers = []
        # add WMS root layer
        wms_layers.append({
            'name': wms_name
        })
        wfs_layers = []
        data_datasets = []
        dataproducts = []
        dataproducts.append(wms_name)
        document_templates = []
        solr_facets = []

        # convert all_services dict to resources list
        all_services = []
        for name, resource in role_permissions.get('all_services', {}).items():
            all_services.append({
                'name': name
            })
            # copy writable flags from all_services to resources_lookup
            if 'writable' in resource:
                lookup = resources_lookup.get(name, {})
                lookup['writable'] = resource['writable']

        # collect permitted resources for role
        role_resources = self.collect_resources(
            all_services, resources_lookup
        )

        # expand to full QWC service permissions
        # NOTE: This generates more permissions than there are actual resources
        #       in a specific service. Any surplus permissions will be ignored.
        for resource in role_resources:
            if 'attributes' in resource:
                # single layer

                # add default 'geometry' column to WMS/WFS attributes
                ogc_attributes = resource['attributes'] + ['geometry']

                # add potential WMS layer
                wms_layers.append({
                    'name': resource['name'],
                    'attributes': ogc_attributes,
                    # NOTE: any info templates are always permitted
                    'info_template': True
                })

                # add potential WFS layer
                wfs_layers.append({
                    'name': resource['name'],
                    'attributes': ogc_attributes
                })

                # add potential Data service dataset
                data_datasets.append({
                    'name': resource['name'],
                    # NOTE: attributes without geometry column
                    'attributes': resource['attributes'],
                    'writable': resource.get('writable', False),
                    # NOTE: always readable
                    'readable': True
                })

                # add potential Solr facet
                solr_facets.append(resource['name'])
            else:
                # add potential group layer to WMS
                wms_layers.append({
                    'name': resource['name']
                })

            dataproducts.append(resource['name'])

            if list(resource.keys()) == ['name']:
                # potential document template has no keys except 'name'
                document_templates.append(resource['name'])

        # add potential internal print layers to WMS
        wms_layers += [
            {'name': name} for name in common_resources
        ]

        # NOTE: assume single WMS service
        full_permissions['wms_services'] = [{
            'name': wms_name,
            'layers': wms_layers,
            # add potential print templates
            'print_templates': common_resources
        }]

        # NOTE: assume single WFS service
        full_permissions['wfs_services'] = [{
            'name': wfs_name,
            'layers': wfs_layers
        }]

        full_permissions['background_layers'] = common_resources
        full_permissions['data_datasets'] = data_datasets
        full_permissions['dataproducts'] = dataproducts
        full_permissions['document_templates'] = document_templates

        # add potential default Solr facets
        solr_facets += common_resources
        full_permissions['solr_facets'] = solr_facets

        return full_permissions

    def collect_resources(self, parent_resources, resources_lookup):
        """Recursively collect resources from 'all_services' and return
        flat list of permitted resources.

        :param list<obj> parent_resources: Parent resources
        :param obj resources_lookup: Lookup for resources with sublayers or
                                     attributes
        """
        resources = []

        for resource in parent_resources:
            # merge with any resource from lookup
            lookup = resources_lookup.get(resource['name'], {})
            resource.update(lookup)

            if 'processed' not in lookup:
                # add resource
                resources.append(resource)
                # mark lookup as processed to skip duplicates
                lookup['processed'] = True
            else:
                # skip duplicates
                continue

            if 'sublayers' in resource:
                # recursively collect sublayers
                sub_resources = [
                    {'name': name} for name in resource['sublayers']
                ]
                resources += self.collect_resources(
                    sub_resources, resources_lookup
                )

        return resources

    # helpers

    def identity_roles(self, identity):
        """Return roles for identity.

        :param obj identity: User identity
        """
        # extract username and group
        username = get_username(identity)
        groups = get_groups(identity)

        # add default public role
        roles = [self.PUBLIC_ROLE_NAME]
        # add any user roles
        roles.extend(self.permissions['users'].get(username, []))
        # add any group roles
        for group in groups:
            roles.extend(self.permissions['groups'].get(group, []))

        # return unique sorted roles
        return sorted(list(set(roles)))

    def resource_permissions(self, resource_key, identity, resource_name=None):
        """Return collected list of resource permissions for identity roles.

        :param str resource_key: Resource key in permissions data
        :param obj identity: User identity
        :param str name: Optional resource name filter
        """
        permissions = []

        roles = self.identity_roles(identity)
        for role in roles:
            # get role permissions
            role_permissions = self.permissions['roles'].get(role, {})
            # get permissions for resource key
            resource_permissions = role_permissions.get(resource_key, {})
            if resource_name is not None:
                # filter by resource name
                for permission in resource_permissions:
                    if isinstance(permission, dict):
                        if permission.get('name') == resource_name:
                            permissions.append(permission)
                    else:
                        if permission == resource_name:
                            permissions.append(permission)
            else:
                permissions.extend(resource_permissions)

        return permissions



from datetime import datetime
import os
import re
from flask import request
from flask.sessions import SecureCookieSessionInterface

from .permissions_reader import PermissionsReader
from .runtime_config import RuntimeConfig


DEFAULT_TENANT = 'default'


class TenantHandlerBase:
    """Tenant handler base class"""

    def __init__(self):
        self.tenant_name = os.environ.get('QWC_TENANT')
        self.tenant_header = os.environ.get('TENANT_HEADER')
        self.tenant_url_re = os.environ.get('TENANT_URL_RE')
        if self.tenant_url_re:
            self.tenant_url_re = re.compile(self.tenant_url_re)

    def is_multi(self):
        return self.tenant_name or self.tenant_header or self.tenant_url_re

    def tenant(self):
        if self.tenant_name:
            return self.tenant_name
        if self.tenant_header:
            return request.headers.get(self.tenant_header, DEFAULT_TENANT)
        if self.tenant_url_re:
            match = self.tenant_url_re.match(request.base_url)
            if match:
                return match.group(1)
            else:
                return DEFAULT_TENANT
        return DEFAULT_TENANT


class TenantHandler(TenantHandlerBase):
    """Tenant handler with configuraton cache"""

    def __init__(self, logger):
        """Constructor

        :param Logger logger: Application logger
        """
        TenantHandlerBase.__init__(self)
        self.logger = logger
        self.handler_cache = {}  # handler_cache[handler_name][tenant]

    def handler(self, service_name, handler_name, tenant):
        """Get service handler for tenant.

        Return None if not yet registered or if config files have changed.

        :param str service_name: Service name
                                 (used for detecting config changes)
        :param str handler_name: Handler name
        :param str tenant: Tenant ID
        """
        handlers = self.handler_cache.get(handler_name)
        if handlers:
            handler = handlers.get(tenant)
            if handler:
                # check for config updates
                last_update = self.last_config_update(service_name, tenant)
                if last_update and last_update < handler.get('last_update'):
                    # cache is up-to-date
                    return handler.get('handler')
                else:
                    # config has changed, remove handler from cache
                    del handlers[tenant]

        return None

    def register_handler(self, handler_name, tenant, handler):
        """Register service handler for tenant"""
        handlers = self.handler_cache.get(handler_name)
        if handlers is None:
            handlers = {}
            self.handler_cache[handler_name] = handlers
        handlers[tenant] = {
            'handler': handler,
            'last_update': datetime.utcnow()
        }
        return handler

    def last_config_update(self, service_name, tenant):
        """Return latest timestamp of config and permission files for a tenant.

        :param str service_name: Service name
        :param str tenant: Tenant ID
        """
        # get latest timestamp of config and permission files
        last_config_update = None
        paths = [
            RuntimeConfig.config_file_path(service_name, tenant),
            PermissionsReader.permissions_file_path(tenant)
        ]
        for path in paths:
            if os.path.isfile(path):
                timestamp = datetime.utcfromtimestamp(
                    os.path.getmtime(path)
                )
                if (
                    last_config_update is None
                    or timestamp > last_config_update
                ):
                    last_config_update = timestamp

        return last_config_update


class TenantPrefixMiddleware:
    """WSGI middleware injecting tenant header in path"""

    def __init__(self, app, _header=None, _ignore_default=None):
        self.app = app
        tenant_header = os.environ.get('TENANT_HEADER')
        if tenant_header:
            self.header = 'HTTP_' + tenant_header.upper()
        else:
            self.header = None
        self.service_prefix = os.environ.get(
            'QWC_SERVICE_PREFIX', '/').rstrip('/') + '/'

    def request_tenant(self, environ):
        if self.header:
            return environ.get(self.header)
        else:
            return None

    def __call__(self, environ, start_response):
        # environ in request http://localhost:9090/base/pages/test.html?arg=1
        # /base is mountpoint (e.g. via WSGIScriptAlias)
        # 'REQUEST_URI': '/base/pages/test.html?arg=1'
        # 'SCRIPT_NAME': '/base'
        # 'PATH_INFO': '/pages/test.html'
        # 'QUERY_STRING': 'arg=1'
        # see also https://www.python.org/dev/peps/pep-3333/#environ-variables
        tenant = self.request_tenant(environ)
        if tenant:
            prefix = self.service_prefix + tenant
            environ['SCRIPT_NAME'] = prefix + environ.get(
                'SCRIPT_NAME', '')
        return self.app(environ, start_response)


class TenantSessionInterface(SecureCookieSessionInterface, TenantHandlerBase):
    """Flask session handler injecting tenant in JWT cookie path"""

    def __init__(self, environ):
        SecureCookieSessionInterface.__init__(self)
        TenantHandlerBase.__init__(self)
        self.service_prefix = environ.get(
            'QWC_SERVICE_PREFIX', '').rstrip('/') + '/'

    def tenant_path_prefix(self):
        """Tenant path prefix /map/org1 ("$QWC_SERVICE_PREFIX/$TENANT")"""
        if self.is_multi():
            return self.service_prefix + self.tenant()
        else:
            return self.service_prefix

    def get_cookie_path(self, app):
        # https://flask.palletsprojects.com/en/1.1.x/api/#flask.sessions.SessionInterface.get_cookie_path
        prefix = self.tenant_path_prefix()
        # Set config as a side effect
        app.config['JWT_ACCESS_COOKIE_PATH'] = prefix
        return prefix




def app_nocache(app):
    """ Adds various cache-disabling headers to all responses returned by the
        application
    :param Flask app: A flask application
    """
    @app.after_request
    def add_header(r):
        r.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
        r.headers["Pragma"] = "no-cache"
        r.headers["Expires"] = "0"
        r.headers['Cache-Control'] = 'public, max-age=0'
        return r



import setuptools

desc = """\
=================
QWC Services Core
=================
Shared modules for QWC services.
"""

setuptools.setup(
    name="qwc-services-core",
    version="1.3.0",
    author="Sourcepole AG",
    author_email="info@sourcepole.ch",
    description="Shared modules for QWC services",
    long_description=desc,
    long_description_content_type="text/x-rst",
    url="https://github.com/qwc-services/qwc-services-core",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)
